from ..model import wtypes
from ..compiler import syscalls, typed_code_static

from ..model.errors import AssertionVMError
from ..model.memory import Ref, Array
from ..model.wtypes import BASE, CONST, VAR, sizeof
from ..model.ops import Dynamic, struct_length, struct_field_types


__all__ = ['builtins', 'log_output', 'on_output']


def wrap_syscall(name):
    syscall = syscalls.index[name]
    def make_body(block):
        block.syscall(syscall, None)
        block.return_()
    param_types = struct_field_types(syscall.in_type)
    return typed_code_static.closure(make_body, param_types, None)


# Map from identifier to dynamic.
builtins = {
    'array0': Dynamic(
        wtypes.type_,
        wtypes.array0,
    ),
    'assert': wrap_syscall('assert'),
    'bool': Dynamic(
        wtypes.type_,
        wtypes.bool_,
    ),
    'dynamic': Dynamic(
        wtypes.type_,
        wtypes.dynamic,
    ),
    'f2i': wrap_syscall('f2i'),
    'f2s': wrap_syscall('f2s'),
    'false': Dynamic.new(
        wtypes.bool_,
        (False,),
    ),
    'float': Dynamic(
        wtypes.type_,
        wtypes.float_,
    ),
    'i2f': wrap_syscall('i2f'),
    'i2s': wrap_syscall('i2s'),
    'inf': Dynamic.new(
        wtypes.float_,
        (float('inf'),),
    ),
    'int': Dynamic(
        wtypes.type_,
        wtypes.int_,
    ),
    'nan': Dynamic.new(
        wtypes.float_,
        (float('nan'),),
    ),
    'never': Dynamic(
        wtypes.type_,
        wtypes.never,
    ),
    'print': wrap_syscall('print'),
    's2f': wrap_syscall('s2f'),
    's2i': wrap_syscall('s2i'),
    'str': Dynamic(
        wtypes.type_,
        wtypes.str_,
    ),
    'true': Dynamic.new(
        wtypes.bool_,
        (True,),
    ),
    'type': Dynamic(
        wtypes.type_,
        wtypes.type_,
    ),
    'version': Dynamic.new(
        wtypes.str_,
        ("0.0.1",),
    ),
    'void': Dynamic(
        wtypes.type_,
        wtypes.void,
    ),
}


# These are useful for the `on_output` argument of `linker.link()`.
on_output = wrap_syscall('on_output')
log_output = wrap_syscall('log_output')
