from . import operators
from ..model import wtypes

from .typed_code import *
from .builtins import *
from ..model.memory import *
from ..model.pack import pack
from ..model.ops import Dynamic


# We need to put dummy location information into some instructions.
# This is the value we will use.
loc = None


def test_main():
    '''
    Builds typed_code hand-compiled from the output of "test_flattener.py"
    (copied at time of writing). The flat code is included in the comments.
    '''
    block = Block()
    """
    (
        ("world", -7),
        new defer fn (int): int
    )
    """
    '''
      ('const_str', 'world', ),
      ('const_int', 7, ),
      ('unary', '-', ),
      ('struct', 2),
      ('capture', 0, ),
      ('cast_to_type', 'the parameter type', False, ),
      ('array', 1, ),
      ('capture', 0, ),
      ('cast_to_type', 'the return type', False, ),
      ('union', 'FUNCTION', 2, ),
      ('cast_to_type', 'SHOULD NOT HAPPEN', False, ),
      ('defer', ),
      ('new', ),
      ('struct', 2),
    '''
    block.defer(wtypes.function([wtypes.int_], wtypes.int_), loc)
    block.new()
    block.static(Dynamic.new(
        wtypes.struct([wtypes.str_, wtypes.int_]),
        ('world', -7)
    ), 1)
    block.struct(2)
    """
    var ((who, (x: int)), j) = <...>
    """
    '''
      ('unpack_struct', 2, ),
      ('reverse', 2, ),
      ('unpack_struct', 2, ),
      ('reverse', 2, ),
      ('init', True, ),
      ('capture', 0, ),
      ('cast', ),
      ('init', True, ),
      ('init', True, ),
    '''
    block.unpack_struct(2, loc)
    block.reverse(2)
    block.unpack_struct(2, loc)
    block.reverse(2)
    block.init(True) # `who`.
    block.cast(wtypes.int_, loc)
    block.init(True) # 'x'.
    block.init(True) # 'j'.

    def loop_body(block):
        """
        {
            print("Hello, "+who+"!");
            // break;
            x -= 1;
            break;
        }
        """

        def then_body(block):
            '''
              ('capture', 1, ),
              ('const_str', 'Hello, ', ),
              ('local', 0, ),
              ('binary', '+', ),
              ('const_str', '!', ),
              ('binary', '+', ),
              ('call', 1, ),
              ('expr', ),
              ('const_int', 1, ),
              ('local', 1, ),
              ('bury', 1, ),
              ('pick', 1, ),
              ('bury', 1, ),
              ('binary', '-', ),
              ('bury', 1, ),
              ('assign', ),
              ('break', ),
            '''
            block.local(0) # `who`.
            block.static(Dynamic.new(wtypes.str_, ('Hello, ',)), 1)
            block.binary(operators.lookup('+', wtypes.str_, wtypes.str_), loc)
            block.static(Dynamic.new(wtypes.str_, ('!',)), 0)
            block.binary(operators.lookup('+', wtypes.str_, wtypes.str_), loc)
            block.static(builtins['print'], 1)
            block.call_function(1, loc)
            block.drop()
            block.local(1) # `x`.
            block.bury(0)
            block.pick(0)
            block.bury(0)
            block.static(Dynamic.new(wtypes.int_, (1,)), 0)
            block.binary(operators.lookup('-', wtypes.int_, wtypes.int_), loc)
            block.swap()
            block.assign(loc)
            block.break_()

        def else_body(block):
            '''
              ('break',),
            '''
            block.break_()

        """
        [implicit "if"] (x>3) <...> [implicit "else { break }"]
        """
        '''
          ('local', 1, ),
          ('const_int', 3, ),
          ('binary', '>', ),
          ('if', 0, [ <then_body> ], [ <else_body> ], ),
        '''
        block.local(1) # `x`.
        block.static(Dynamic.new(wtypes.int_, (3,)), 0)
        block.binary(operators.lookup('>', wtypes.int_, wtypes.int_), loc)
        block.if_(0, then_body, else_body, loc)

    """
    while <...>
    """
    '''
      ('loop', [ <loop_body> ], ),
    '''
    block.loop(loop_body)
    # TODO: The `factorial` function is static. Move elsewhere.

    def factorial_body(block):
        """
        fn dec() { return n - 1; }
        if (n == 0) {
          return 1;
        } else {
          return n * factorial(dec(n));
        }
        """
        block.unpack_struct(1, loc)
        block.init(False) # `n`.

        def dec_body(block):
            """
            return n - 1
            """
            '''
              ('capture', 0, ),
              ('const_int', 1, ),
              ('binary', '-', ),
              ('return', ),
            '''
            block.unpack_struct(0, loc)
            block.capture(0) # `n`.
            block.static(Dynamic.new(wtypes.int_, (1,)), 0)
            block.binary(operators.lookup('-', wtypes.int_, wtypes.int_), loc)
            block.return_()

        """
        fn dec() <...>
        """
        '''
          ('array', 0, ),
          ('local', 0, ),
          ('closure', [ <dec_body> ], [], 1, ),
          ('init', False, ),
        '''
        # Static param types omitted.
        block.local(0) # `n`.
        block.closure(
            dec_body,
            Array.from_list([], wtypes.sizeof(wtypes.type_)),
            1,
            loc,
        )
        block.init(False) # `dec`.

        def then_body(block):
            """
            return 1  [with implicit cast to int]
            """
            '''
              ('const_int', 1, ),
              ('capture', 0, ),
              ('cast', ),
              ('return', ),
            '''
            block.static(Dynamic.new(wtypes.int_, (1,)), 0)
            block.return_()

        def else_body(block):
            """
            return n * factorial(dec())  [with implicit cast to int]
            """
            '''
              ('local', 0, ),
              ('capture', 1, ),
              ('local', 1, ),
              ('call', 0, ),
              ('call', 1, ),
              ('binary', '*', ),
              ('capture', 0, ),
              ('cast', ),
              ('return', ),
            '''
            block.local(0) # `n`.
            block.capture(0) # `factorial`.
            block.local(1) # `dec`.
            block.call_function(0, loc)
            block.call_function(1, loc)
            block.binary(operators.lookup('*', wtypes.int_, wtypes.int_), loc)
            block.cast(wtypes.int_, loc)
            block.return_()

        """
        if (n == 0) <...> else <...>
        """
        '''
          ('local', 0, ),
          ('const_int', 0, ),
          ('compare', '==', ),
          ('if', 0, [ <then_body> ], [ <else_body> ], ),
        '''
        block.local(0) # `n`.
        block.static(Dynamic.new(wtypes.int_, (0,)), 0)
        block.binary(operators.lookup('==', wtypes.int_, wtypes.int_), loc)
        block.if_(0, then_body, else_body, loc)

    """
    fn factorial(n: int): int <...>
    """
    '''
      ('capture', 0, ),
      ('cast_to_type', 'the parameter type', True, ),
      ('array', 1, ),
      ('capture', 0, ),
      ('cast_to_type', 'the return type', True, ),
      ('pick', 1, ),
      ('pick', 1, ),
      ('union', 'FUNCTION', 2, ),
      ('defer', ),
      ('bury', 2, ),
      ('pick', 2, ),
      ('closure', [ <factorial_body> ], [VAR, ], 2, ),
      ('pick', 1, ),
      ('fulfil', ),
      ('init', False, ),
    '''
    block.defer(wtypes.function([wtypes.int_], wtypes.int_), loc)
    # Static param types removed.
    # Static return type removed.
    # Two static captures removed.
    block.pick(0)
    block.closure(
        factorial_body,
        Array.from_list(
            [wtypes.int_.read()],
            wtypes.sizeof(wtypes.type_),
        ),
        1,
        loc,
    )
    block.yzy()
    block.fulfil(loc)
    block.init(False) # `factorial`.
    # TODO: The `like_to_like` function is static. Move elsewhere.

    def like_to_like_body(block):
        """
        return fn (t): t  [with implicit cast to type]
        """
        '''
          ('local', 0, ),
          ('cast_to_type', 'the parameter type', False, ),
          ('array', 1, ),
          ('local', 0, ),
          ('cast_to_type', 'the return type', False, ),
          ('union', 'FUNCTION', 2, ),
          ('cast_to_type', 'SHOULD NOT HAPPEN', False, ),
          ('return', ),
        '''
        block.unpack_struct(1, loc)
        block.init(False) # `t`.
        block.local(0) # `t`.
        block.cast(wtypes.type_, loc)
        block.array(wtypes.type_, 1, loc)
        block.local(0) # `t`.
        block.cast(wtypes.type_, loc)
        block.union('FUNCTION', 2)
        block.cast(wtypes.type_, loc)
        block.return_()

    """
    fn like_to_like(t: type) <...>
    """
    '''
      ('capture', 2, ),
      ('cast_to_type', 'the parameter type', True, ),
      ('array', 1, ),
      ('closure', [ <like_to_like_body> ], [2, ], 0, ),
      ('init', False, ),
    '''
    # This should really be static, but let's pretend it's not.
    block.closure(
        like_to_like_body,
        Array.from_list(
            [wtypes.type_.read()],
            wtypes.sizeof(wtypes.type_),
        ),
        0,
        loc,
    )
    block.init(False) # `like_to_like`.


def test_non_returning_function():
    block = Block()
    """
    Roughly fn() { while true {} }

    Interesting because it contains no 'return' instructions.
    """
    def body(block):
        block.unpack_struct(0, loc)
        block.loop(lambda block: None)
    block.closure(
        body,
        Array.from_list([], wtypes.sizeof(wtypes.type_)),
        0,
        loc,
    )
