'''
The constness of a location describes who is allowed to change its value:

 - If a location is `CONST`, you may assume that its value will not change.
 - If a location is `VAR`, you may change its value.
 - If a location is `BASE`, you may do neither.

Constnesses are represented as Welly union values. The recommended way to
deconstruct a constness (i.e. to test its value) is to use the `switch()`
function defined here.
'''


from . import pack


__all__ = [
    'BASE', 'CONST', 'VAR', 'switch', 'check_valid',
]


BASE = pack.tag('BASE')
CONST = pack.tag('CONST')
VAR = pack.tag('VAR')


_SWITCH_DICT = {'BASE': '', 'CONST': '', 'VAR': ''}

def switch(constness):
    '''
    Returns the tag of the constness, i.e. 'BASE', 'CONST' or 'VAR'.

    This is the recommended way of deconstructing (i.e. testing) a constness.

     - constness - a Ref to a constness.
    '''
    tag, _ = pack.switch(constness, _SWITCH_DICT)
    return tag


def check_valid(constness):
    _ = switch(constness)


def join(c, d):
    '''
    Returns `c` if it is equal to `d`, otherwise returns `BASE`.

     - c - A Ref to a constness.
     - d - A Ref to a constness.
    '''
    if switch(c) == switch(d):
        return c
    else:
        return BASE
