import pytest

from . import errors

from .memory import Ref
from .defer import defer, fulfil, read


# A deferrable value. (It's a Welly union value).
a = ('a', [])
a_ref = Ref.new(a)


def test_main():
    tag, ptr = defer().read()
    assert tag is None
    assert type(ptr) is list
    assert len(ptr) == 2
    assert ptr[0] is None
    assert ptr[1] is ptr

    x = defer()
    fulfil(x, a_ref)
    assert read(x) == a

    x = defer()
    y = x.copy()
    fulfil(y, a_ref)
    assert read(x) == a
    assert read(y) == a

    x = defer()
    y = defer()
    fulfil(x, y)
    fulfil(y, a_ref)
    assert read(x) == a
    assert read(y) == a

    x = defer()
    fulfil(x, x)
    fulfil(x, a_ref)
    assert read(x) == a


def test_read_deferred():
    x = defer()
    with pytest.raises(errors.UnfulfilledValueVMError):
        read(x)

    x = defer()
    y = defer()
    fulfil(x, y)
    with pytest.raises(errors.UnfulfilledValueVMError):
        read(x)


def test_double_fulfil():
    x = defer()
    fulfil(x, a_ref)
    with pytest.raises(errors.FulfilledValueVMError):
        fulfil(x, a_ref)

    x = defer()
    y = x.copy()
    fulfil(x, a_ref)
    with pytest.raises(errors.FulfilledValueVMError):
        fulfil(y, a_ref)

    x = defer()
    y = x.copy()
    fulfil(x, y)
    fulfil(y, a_ref)
    with pytest.raises(errors.FulfilledValueVMError):
        fulfil(x, a_ref)
