from . import defer, pack

from welly.printer import pretty

from ..compiler.builtins import builtins

from .pretty_type import wrap
from .const import BASE, CONST, VAR
from .wtypes import *

def test_main():
    assert pretty(wrap(bool_)) == 'bool'
    assert pretty(wrap(float_)) == 'float'
    assert pretty(wrap(int_)) == 'int'
    assert pretty(wrap(str_)) == 'str'
    assert pretty(wrap(ref(BASE, int_))) == 'ref int'
    assert pretty(wrap(ref(CONST, int_))) == 'ref const int'
    assert pretty(wrap(ref(VAR, int_))) == 'ref var int'
    assert pretty(wrap(array0)) == 'array0'
    assert pretty(wrap(array(BASE, int_))) == 'array int'
    assert pretty(wrap(array(CONST, int_))) == 'array const int'
    assert pretty(wrap(array(VAR, int_))) == 'array var int'
    assert pretty(wrap(struct([]))) == 'void'
    assert pretty(wrap(struct([int_]))) == 'struct(int)'
    assert pretty(wrap(struct([int_, str_]))) == 'struct(int, str)'
    assert pretty(wrap(
        union([])
    )) == 'union {}'
    assert pretty(wrap(
        union([('FOO', [])])
    )) == 'union {FOO()}'
    assert pretty(wrap(
        union([('FOO', []), ('BAR', [int_, str_])])
    )) == 'union {FOO(); BAR(int, str)}'
    assert pretty(wrap(
        function([], void)
    )) == 'fn(): (void)'
    assert pretty(wrap(
        function([int_], bool_)
    )) == 'fn(int): (bool)'
    assert pretty(wrap(
        function([int_, str_], float_)
    )) == 'fn(int, str): (float)'
    assert pretty(wrap(dynamic)) == 'dynamic'


def test_defer():
    assert pretty(wrap(defer.defer())) == 'unfulfilled'


def test_type():
    assert pretty(wrap(type_)) == 'type'


def test_builtins():
    assert pretty(wrap(builtins['never'].ref)) == 'union {}'
    assert pretty(wrap(builtins['void'].ref)) == 'void'
    assert pretty(wrap(builtins['type'].ref)) == 'type'
