from . import defer, pack

from welly.printer import pretty

from ..compiler.builtins import builtins

from .pretty_value import wrap
from .memory import Ref, Array
from .const import BASE, CONST, VAR
from .wtypes import *
from .ops import Dynamic


def test_main():
    assert pretty(wrap(Dynamic.new(int_, (5,)))) == '5'
    assert pretty(wrap(Dynamic.new(float_, (5.0,)))) == '5.0'
    assert pretty(wrap(Dynamic.new(str_, ('h',)))) == '"h"'
    assert pretty(wrap(Dynamic.new(
        ref(CONST, int_),
        (Ref.new((1,)),),
    ))) == '&1'
    assert pretty(wrap(Dynamic.new(
        array0,
        (Ref.new(()), 0),
    ))) == '[]'
    assert pretty(wrap(Dynamic.new(
        array(CONST, int_),
        (Ref.new((1, 2, 3)), 3),
    ))) == '[1, 2, 3]'
    assert pretty(wrap(Dynamic.new(
        struct([int_, str_]),
        (5, 'h'),
    ))) == '(5, "h")'
    assert pretty(wrap(Dynamic(
        union([('FOO', [int_])]),
        pack.tag('FOO', 5),
    ))) == 'FOO(5)'
    assert pretty(wrap(Dynamic.new(
        function([], void),
        (lambda _: Ref.new(()), []),
    ))) == '<fn>'
    assert pretty(wrap(Dynamic.new(
        dynamic,
        Dynamic.new(int_, (5,)).to_welly(),
    ))) == 'box 5'


def test_defer():
    assert pretty(wrap(Dynamic(
        union([('FOO', [int_])]),
        defer.defer(),
    ))) == 'unfulfilled'
    assert pretty(wrap(Dynamic(
        function([], void),
        defer.defer(),
    ))) == 'unfulfilled'


def test_type():
    assert pretty(wrap(Dynamic(type_, int_))) == 'int'


def test_builtins():
    assert pretty(wrap(builtins['true'])) == 'true'
    assert pretty(wrap(builtins['false'])) == 'false'
    assert pretty(wrap(builtins['print'])) == '<fn>'
