from .repl import *
from ..model.errors import *
from ..model.wtypes import int_, str_, void


def test_describe_exception():
    loc = None
    assert describe_exception(TypeVMError(int_, str_, None, loc)) == \
        "TypeVMError: value of type int is not of type str"
    assert describe_exception(TypeVMError(int_, str_, 'frog', loc)) == \
        "TypeVMError: frog is not of type str"
    assert describe_exception(
        TypeInferenceVMError(int_, str_, 'subtype', 'the example type', loc)
    ) == (
        "TypeInferenceVMError: cannot compute the example type; "
        "int and str have no common subtype"
    )
    assert describe_exception(CallTypeVMError(int_, loc)) == \
        "CallTypeVMError: cannot call values of type int"
    assert describe_exception(DeferTypeVMError(int_, loc)) == \
        'DeferTypeVMError: cannot "defer" or "fulfil" values of type int'
    assert describe_exception(DereferenceTypeVMError(int_, loc)) == \
        "DereferenceTypeVMError: cannot dereference values of type int"
    assert describe_exception(SubscriptTypeVMError(int_, loc)) == \
        "SubscriptTypeVMError: cannot subscript values of type int"
    assert describe_exception(SwitchTypeVMError(int_, loc)) == \
        'SwitchTypeVMError: cannot "switch" on values of type int'
    assert describe_exception(UnpackTypeVMError(int_, loc)) == \
        "UnpackTypeVMError: cannot unpack values of type int"
    assert describe_exception(StructIndexVMError(void, 1, loc)) == \
        "StructIndexVMError: index 1 is out of range for type void"
    assert describe_exception(FieldAccessVMError(int_, 'x', loc)) == \
        "FieldAccessVMError: type int has no field 'x'"
    assert describe_exception(BinaryOperatorVMError('+', int_, void, loc)) == \
        'BinaryOperatorVMError: binary operator "+" is not defined for ' \
        '(int, void)'
    assert describe_exception(UnaryOperatorVMError('-', void, loc)) == \
        'UnaryOperatorVMError: unary operator "-" is not defined for void'


def test_repl():
    run_file('')
    run_file('1')
    run_file('\xf4')
