from . import code_types, core

from ..model import wtypes, pack, defer
from ..model.memory import Ref, Array


def _types_to_welly(types):
    return Array.from_list(
        [defer.read(t) for t in types],
        wtypes.sizeof(wtypes.type_),
    )

def _registers_to_welly(registers):
    return Array.from_list(
        [r.to_welly() for r in registers],
        wtypes.sizeof(code_types.register),
    )


class Compiler(object):
    '''
    Represents a partially constructed array const op.

    Public fields:
     - bindings - a list of Registers representing the result registers of the
       previous instruction, or `None` if the previous instruction does not
       fall through.
    '''
    def __init__(self, bindings):
        self._code = [] # tuples, representing values of type `op`.
        self.bindings = bindings

    def get_code(self):
        '''
        Returns an Array.
        '''
        assert self.bindings is None
        return Array.from_list(self._code, wtypes.sizeof(code_types.op))

    def _append(self, op, results=None):
        assert self.bindings is not None
        self._code.append(
            pack.pack(_registers_to_welly(self.bindings), defer.read(op)))
        self.bindings = results

    def arithmetic(self, intrinsic, operands, results):
        operands = _registers_to_welly(operands)
        self._append(
            pack.tag('ARITHMETIC', instrinsic, operands),
            results,
        )

    def goto(self, target, arguments):
        arguments = _registers_to_welly(arguments)
        self._append(
            pack.tag('GOTO', target.to_welly(), arguments),
        )

    def continuation(self, arg_types, body, result):
        arg_types = _types_to_welly(arg_types)
        assert isinstance(body, Array)
        self._append(
            pack.tag('CONTINUATION', arg_types, body),
            [result],
        )

    def function(self, arg_types, body, result):
        arg_types = _types_to_welly(arg_types)
        assert isinstance(body, Array)
        self._append(
            pack.tag('FUNCTION', arg_types, body),
            [result],
        )

    def switch(self, selector, case_name, case_body, else_body):
        self._append(
            pack.tag('SWITCH', selector, case_name, case_body, else_body),
        )
