import pytest

from . import code_types, core

from ..model import pack
from .code_types import R0, R1, R2, R3, R4, R5, R6, R7, R8, R9
from .compiler import *


def test_undercomplete():
    with pytest.raises(AssertionError):
        c = Compiler([R0])
        c.get_code()


def test_overcomplete():
    with pytest.raises(AssertionError):
        c = Compiler([R0])
        c.goto(R0, [])
        c.goto(R0, [])


def test_goto():
    c = Compiler([R0, R1, R2])
    c.goto(R0, [R1, R2])
    a = c.get_code()
    (op_ref,) = a.to_list()
    bindings, instruction = pack.unpack(op_ref, 'a1r2')
    print(bindings.to_list())
    (r0_ref, r1_ref, r2_ref) = bindings.to_list()
    assert r0_ref.read() == (0,)
    assert r1_ref.read() == (1,)
    assert r2_ref.read() == (2,)
    tag, payload = code_types.switch(instruction)
    assert tag == 'GOTO'
    (r0, arguments) = payload
    assert r0 == 0
    (r1_ref, r2_ref) = arguments.to_list()
    assert r1_ref.read() == (1,)
    assert r2_ref.read() == (2,)
